/***********************************************************************************************************************
 *
 * blueBill Mobile - Android - open source birding
 * Copyright (C) 2009-2011 by Tidalwave s.a.s. (http://www.tidalwave.it)
 *
 ***********************************************************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 *
 ***********************************************************************************************************************
 *
 * WWW: http://bluebill.tidalwave.it/mobile
 * SCM: https://kenai.com/hg/bluebill~android-src
 *
 **********************************************************************************************************************/
package it.tidalwave.bluebill.mobile.news;

import javax.annotation.CheckForNull;
import javax.annotation.Nonnegative;
import javax.annotation.Nonnull;
import javax.inject.Provider;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.Set;
import java.util.TreeSet;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.MalformedURLException;
import java.net.URL;
import org.openide.util.lookup.ServiceProvider;
import it.tidalwave.util.Id;
import it.tidalwave.util.NotFoundException;
import it.tidalwave.util.logging.Logger;
import it.tidalwave.util.thread.ThreadAssertions;
import it.tidalwave.util.thread.ThreadType;
import it.tidalwave.util.thread.annotation.ThreadConfined;
import it.tidalwave.netbeans.util.Locator;
import it.tidalwave.mobile.io.FileSystem;
import it.tidalwave.mobile.io.IoUtils;
import it.tidalwave.mobile.io.MasterFileSystem;
import it.tidalwave.mobile.io.StringsIO;
import it.tidalwave.semantic.document.Document;
import it.tidalwave.mobile.rss.RSSFeed;
import it.tidalwave.mobile.rss.RSSFeedReader;
import it.tidalwave.bluebill.mobile.network.NetworkingPreferences;
import lombok.Cleanup;
import static it.tidalwave.mobile.rss.RssVocabulary.*;

/***********************************************************************************************************************
 *
 * 
 * @stereotype Service
 * 
 * @author  Fabrizio Giudici
 * @version $Id$
 *
 **********************************************************************************************************************/
@ServiceProvider(service=NewsService.class)
public class DefaultNewsService implements NewsService
  {
    private static final String CLASS = DefaultNewsService.class.getName();
    private static final Logger logger = Logger.getLogger(CLASS);
    private static final Class<DefaultNewsService> _ = DefaultNewsService.class;

    private static final String NEWS_FEED_URL = "http://bluebill.tidalwave.it/mobile/blog.rss";

    private static final long ONE_DAY = 24 * 60 * 60 * 1000;
    private static final long ONE_MINUTE = 60 * 1000;

    protected enum Status
      {
        NEEDS_DOWNLOAD,
        DOWNLOADING,
        DOWNLOADED
      }

    @Nonnegative
    protected long refreshPeriod = ONE_DAY;

    @Nonnull
    protected Status status = Status.NEEDS_DOWNLOAD;

    @CheckForNull
    protected Thread thread;

    @Nonnull
    protected final File cachedFile;

    @Nonnull
    protected final File readLogFile;

    @CheckForNull
    protected RSSFeed rssFeed;

    @Nonnull
    private final Provider<MasterFileSystem> masterFileSystem = Locator.createProviderFor(MasterFileSystem.class);

    @Nonnull
    private final Provider<NetworkingPreferences> networkingPreferences = Locator.createProviderFor(NetworkingPreferences.class);

    @CheckForNull
    protected Thread downloaderThread;

    private final Set<Id> readMessages = new TreeSet<Id>();

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @ThreadConfined(type=ThreadType.NOT_UI)
    public DefaultNewsService()
      throws IOException
      {
        ThreadAssertions.assertThread(ThreadType.NOT_UI);
        final FileSystem internalFileSystem = masterFileSystem.get().getInternalFileSystem();
        cachedFile = internalFileSystem.getFile("blog.rss");
        readLogFile = internalFileSystem.getFile("blog-read-messages.txt");
        loadReadMessagesLog();
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override
    public void start()
      {
        logger.info("start()");
        
        if (thread == null)
          {
            if (!networkingPreferences.get().isNetworkConnectionAllowed())
              {
                logger.info(">>>> not checking, internet connections are disabled");
              }
            else
              {
                thread = new Thread()
                  {
                    @Override
                    public void run()
                      {
                        try
                          {
                            updateCache();
                          }
                        catch (Exception e)
                          {
                            logger.warning("While running news feed check: %s", e);
                            logger.throwing(CLASS, "check()", e);
                          }

                        thread = null;
                      }
                  };

                thread.start();
              }
          }
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override @Nonnull
    public void getNewsFeed (final @Nonnull NewsFeedCallback callBack)
      {
        logger.info("getNewsFeed(%s)", callBack);

        if (rssFeed != null)
          {
            callBack.notifyFeedAvailable(rssFeed);
          }
        else
          {
            if (status == Status.NEEDS_DOWNLOAD)
              {
                try
                  {
                    computeCacheStatus();
                  }
                catch (Exception e)
                  {
                    callBack.notifyFeedUnavailable();
                    return;
                  }
              }

            if (status != Status.NEEDS_DOWNLOAD)
              {
                loadNewsFeedFromCache(callBack);
              }
            // NOT_DOWNLOADED
            else if (networkingPreferences.get().isNetworkConnectionAllowed())
              {
                downloadNewsAndLoadRssFeed(callBack);
              }
            else
              {
                callBack.notifyFeedCouldBeDownloaded();
              }
          }
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override 
    public void downloadNewsFeedInBackground (final @Nonnull NewsFeedCallback callBack) 
      {
        downloaderThread = new Thread()
          {
            @Override
            public void run()
              {
                downloadNewsAndLoadRssFeed(callBack);
                downloaderThread = null;
              }
          };

        downloaderThread.start();
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override 
    public void markRead (final @Nonnull Document message)
      throws IOException
      {
        logger.info("markRead(%s)", message);
        readMessages.add(message.getId());
        storeReadMessagesLog();
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override 
    public void markAllRead (final @Nonnull RSSFeed rssFeed)
      throws IOException
      {
        logger.info("markAllRead(%s)", rssFeed);

        try
          {
            for (final Document message : rssFeed.get(MESSAGES))
              {
                readMessages.add(message.getId());
              }

            storeReadMessagesLog();
          }
        catch (NotFoundException e)
          {
            logger.warning("markAllRead(): %s", e);
            logger.throwing(CLASS, "markAllRead()", e);
          }
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override
    public boolean isRead (final @Nonnull Document message)
      {
        return readMessages.contains(message.getId());
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    protected void downloadNewsAndLoadRssFeed (final @Nonnull NewsFeedCallback callBack)
      {
        logger.fine("downloadNewsAndLoadRssFeed(%s)", callBack);

        try
          {
            downloadNewsFeed();
            loadNewsFeedFromCache(callBack);
          }
        catch (Exception e)
          {
            callBack.notifyFeedUnavailable();
          }
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    protected void loadNewsFeedFromCache (final @Nonnull NewsFeedCallback callBack)
      {
        try
          {
            synchronized (this)
              {
                while (status != Status.DOWNLOADED)
                  {
                    wait();
                  }
              }

            loadNewsFeedFromCache();
            callBack.notifyFeedAvailable(rssFeed);
          }
        catch (Exception e)
          {
            callBack.notifyFeedUnavailable();
          }
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    protected void updateCache()
      throws Exception
      {
        logger.info("updateCache()");
        computeCacheStatus();

        if (status == Status.NEEDS_DOWNLOAD)
          {
            downloadNewsFeed();
            loadNewsFeedFromCache();
            eventuallyPostUnreadNewsNotification();
          }
      }

    /*******************************************************************************************************************
     *
     * Loads the news feed from the cache.
     *
     ******************************************************************************************************************/
    protected void loadNewsFeedFromCache()
      throws Exception, FileNotFoundException
      {
        @Cleanup final InputStream is = new FileInputStream(cachedFile);
        final RSSFeedReader reader = new RSSFeedReader();
        rssFeed = reader.readMessages(is);
      }

    /*******************************************************************************************************************
     *
     * COmputes the cache status. The status is updated as:
     * 
     * <ul>
     * <li>{@code NEEDS_DOWNLOAD} if the cache is empty or outdated;</li>
     * <li>{@code DOWNLOADED} if the cache is not empty and updated.</li>
     * </ul>
     *
     ******************************************************************************************************************/
    protected void computeCacheStatus()
      throws Exception
      {
        logger.info("computeCacheStatus()");
        logger.info(">>>> cached file is %s", cachedFile.getAbsolutePath());

        if (!cachedFile.exists())
          {
            logger.info(">>>> cached file doesn't exist, needs download");
            setStatus(Status.NEEDS_DOWNLOAD);
          }
        else
          {
            final long lastModified = cachedFile.lastModified();
            final long now = getTimestamp();
            final long delta = now - lastModified;

            setStatus((delta >= refreshPeriod) ? Status.NEEDS_DOWNLOAD : Status.DOWNLOADED);

            logger.info(">>>> now: %s, latest news download: %s, delta: %d min, status: %s",
                        new Date(now), new Date(lastModified), delta / ONE_MINUTE, status);
          }
      }

    /*******************************************************************************************************************
     *
     * Downloads the news in the current thread, blocking until they are ready.
     *
     ******************************************************************************************************************/
    protected void downloadNewsFeed()
      throws FileNotFoundException, IOException, MalformedURLException
      {
        logger.info("downloadNewsFeed()");
        setStatus(Status.DOWNLOADING);
        // TODO: be smart, check the last update time of the feed
        @Cleanup final InputStream is = createUrl(NEWS_FEED_URL).openStream();
        @Cleanup final OutputStream os = new FileOutputStream(cachedFile);
        IoUtils.copy(is, os);
        setStatus(Status.DOWNLOADED);
      }

    /*******************************************************************************************************************
     *
     * Checks whether there are unread news and eventually posts a related notification.
     *
     ******************************************************************************************************************/
    protected void eventuallyPostUnreadNewsNotification()
      {
        logger.info("checkNotification()");

        try
          {
            final List<Id> messageIds = new ArrayList<Id>();

            for (final Document message : rssFeed.get(MESSAGES))
              {
                messageIds.add(message.getId());
              }

            messageIds.removeAll(readMessages);

            logger.info(">>>> unread messages: %d", messageIds.size());
            
            if (!messageIds.isEmpty())
              {
                postUnreadNewsNotification();
              }
          }
        catch (NotFoundException e)
          {
            logger.warning("markAllRead(): %s", e);
            logger.throwing(CLASS, "markAllRead()", e);
          }
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    protected void postUnreadNewsNotification()
      {
      }

    /*******************************************************************************************************************
     *
     * In a specific method so it can be mocked.
     * 
     * TODO: use a TimeProvider
     *
     ******************************************************************************************************************/
    protected long getTimestamp()
      {
        return System.currentTimeMillis();
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Nonnull
    protected URL createUrl (final String string)
      throws IOException
      {
        return new URL(string);
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    protected synchronized void setStatus (final @Nonnull Status status)
      {
        logger.info("setStatus(%s)", status);
        this.status = status;
        notifyAll();
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    private void loadReadMessagesLog()
      throws IOException
      {
        if (readLogFile.exists())
          {
            readMessages.addAll(new StringsIO(readLogFile).load(new StringsIO.Processor<Id>()
              {
                @Nonnull
                public Id processString (final @Nonnull String string)
                  {
                    return new Id(string);
                  }
              }));
          }
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    private void storeReadMessagesLog()
      throws IOException
      {
        new StringsIO(readLogFile).store(readMessages);
      }
  }
