/***********************************************************************************************************************
 *
 * blueBill Mobile - Android - open source birding
 * Copyright (C) 2009-2011 by Tidalwave s.a.s. (http://www.tidalwave.it)
 *
 ***********************************************************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 *
 ***********************************************************************************************************************
 *
 * WWW: http://bluebill.tidalwave.it/mobile
 * SCM: https://kenai.com/hg/bluebill~android-src
 *
 **********************************************************************************************************************/
package it.tidalwave.bluebill.mobile.observation.report;

import it.tidalwave.semantic.document.Document;
import javax.annotation.Nonnull;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Date;
import java.util.List;
import java.util.Collection;
import java.util.Locale;
import java.util.ResourceBundle;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import org.openide.util.NbBundle;
import it.tidalwave.util.AsException;
import it.tidalwave.util.NotFoundException;
import it.tidalwave.netbeans.util.Locator;
import it.tidalwave.geo.Coordinate;
import it.tidalwave.geo.Range;
import it.tidalwave.observation.Cardinality;
import it.tidalwave.observation.Location;
import it.tidalwave.observation.Observation;
import it.tidalwave.observation.ObservationItem;
import it.tidalwave.observation.ObservationSet;
import it.tidalwave.bluebill.mobile.observation.BirdGender;
import it.tidalwave.bluebill.mobile.observation.ObservationType;
import it.tidalwave.bluebill.mobile.taxonomy.TaxonomyPreferences;
import it.tidalwave.bluebill.taxonomy.mobile.Taxon;
import static it.tidalwave.observation.TextNote.TextNote;
import static it.tidalwave.bluebill.mobile.observation.ObservationType.*;
import static it.tidalwave.bluebill.mobile.observation.BirdGender.BirdGender; // it _is_ needed
import static it.tidalwave.bluebill.mobile.observation.BirdGender.*;
import static it.tidalwave.bluebill.taxonomy.mobile.Taxon.Taxon;

/***********************************************************************************************************************
 *
 * @author  Fabrizio Giudici
 * @version $Id$
 *
 **********************************************************************************************************************/
public class CSVReportGenerator extends ReportGenerator
  {
    private Locale primaryTaxonLocale;
    
    private final List<String> strings = new ArrayList<String>();

    private String prefix = "";
    
    private String separator = ";";

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    public CSVReportGenerator() 
      {
        super("CSV", Document.CSV_MIME_TYPE);
      }
    
    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Override
    public void preVisit (final @Nonnull ObservationSet observationSet)
      {
        primaryTaxonLocale = Locator.find(TaxonomyPreferences.class).getTaxonomyLocales().get(0);
        buffer.append(NbBundle.getMessage(CSVReportGenerator.class, "csvHeader").replace("$", separator) + "\n");
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Override
    public void preVisit (final @Nonnull Observation observation)
      {
        final Date date = observation.getDate();
        final Location location = observation.getLocation();

        final StringBuilder buffer = new StringBuilder();
        final DateFormat dateFormat = new SimpleDateFormat("yyyy/MM/dd");
        final DateFormat timeFormat = new SimpleDateFormat("HH:mm");
        buffer.append(dateFormat.format(date)).append(separator);
        buffer.append(timeFormat.format(date)).append(separator);
        buffer.append(location.getDisplayName()).append(separator);

        try
          {
            final Range range = location.as(Range);
            final Coordinate coordinate = range.getCoordinate();
            final int precision = (int)Math.round(range.getRadius());
            buffer.append(String.format(Locale.getDefault(), "%1$.5f%4$s%2$.5f%4$s%3$d", coordinate.getLatitude(), 
                                                                                         coordinate.getLongitude(), 
                                                                                         precision, 
                                                                                         separator));
          }
        catch (AsException e)
          {
            buffer.append(separator).append(separator); // Range not available, never mind
          }

        prefix = buffer.toString();
        strings.clear();
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Override
    public void visit (final @Nonnull ObservationItem observationItem)
      {
        final StringBuilder buffer = new StringBuilder();
        final Taxon taxon = observationItem.getObservable().as(Taxon);
        final Cardinality cardinality = observationItem.getCardinality();
        buffer.append(prefix).append(separator);
        
        try 
          {
            buffer.append(taxon.getDisplayName(primaryTaxonLocale)).append(separator);
          } 
        catch (NotFoundException e) 
          {
            buffer.append("ERROR").append(separator);
          }
        
        buffer.append(taxon.getScientificName()).append(separator);
        
        if (!cardinality.equals(Cardinality.UNDEFINED))
          {
            buffer.append(cardinality);
          }
                
        buffer.append(separator);
        
        try
          {
            final BirdGender gender = observationItem.getObservable().as(BirdGender);

            if (!gender.equals(NOT_RECORDED))
              {
                // FIXME: why ResourceBundler and not NbBundle?
                final ResourceBundle bundle = ResourceBundle.getBundle("it/tidalwave/bluebill/mobile/observation/Bundle", primaryTaxonLocale);
                final String shortGender = bundle.getString("shortGender_" + gender.toString());
                buffer.append(shortGender.trim());
              }            
          }
        catch (AsException e)
          {
            // ok, no gender  
          }
        
        buffer.append(separator);  

        final Collection<? extends ObservationType> observationTypes = 
                observationItem.getObservable().getLookup().lookupAll(ObservationType.class);

        for (final ObservationType observationType : Arrays.asList(WATCHED, LISTENED))
          {
            if (observationTypes.contains(observationType))
              {
                buffer.append(observationType.getDisplayName(primaryTaxonLocale));
              }        

            buffer.append(separator);
          }
        
        try
          {
            buffer.append(observationItem.getObservable().as(TextNote).stringValue());
          }
        catch (AsException e)
          {
            // ok, no text note
          }

        strings.add(buffer.toString());
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Override
    public void postVisit (final @Nonnull Observation observation)
      {
        Collections.sort(strings);

        for (final String string : strings)
          {
            buffer.append(string).append("\n");
          }

        strings.clear();
      }
  }
