/***********************************************************************************************************************
 *
 * blueBill Mobile - Android - open source birding
 * Copyright (C) 2009-2011 by Tidalwave s.a.s. (http://www.tidalwave.it)
 *
 ***********************************************************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 *
 ***********************************************************************************************************************
 *
 * WWW: http://bluebill.tidalwave.it/mobile
 * SCM: https://kenai.com/hg/bluebill~android-src
 *
 **********************************************************************************************************************/
package it.tidalwave.bluebill.mobile.observation.ui;

import javax.annotation.Nonnull;
import javax.inject.Provider;
import java.util.Date;
import java.util.Calendar;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.io.File;
import java.io.IOException;
import it.tidalwave.util.logging.Logger;
import it.tidalwave.util.thread.ThreadAssertions;
import it.tidalwave.util.thread.ThreadType;
import it.tidalwave.netbeans.util.Locator;
import it.tidalwave.observation.Observation;
import it.tidalwave.observation.Observation.Builder;
import it.tidalwave.observation.ObservationItem;
import it.tidalwave.observation.ObservationManager;
import it.tidalwave.observation.ObservationSet;
import it.tidalwave.mobile.util.DateUpdater;
import it.tidalwave.mobile.io.FileSystem;
import it.tidalwave.mobile.io.MasterFileSystem;
import it.tidalwave.mobile.location.LocationFinder;
import it.tidalwave.mobile.ui.QuestionWithFeedback;
import it.tidalwave.semantic.document.Document;
import it.tidalwave.semantic.document.DocumentWriter;
import it.tidalwave.bluebill.mobile.observation.ObservationClipboard;
import it.tidalwave.bluebill.mobile.observation.SharingPreferences;
import it.tidalwave.bluebill.taxonomy.mobile.Taxon;
import static org.openide.util.NbBundle.*;
import static it.tidalwave.role.StringRenderable.StringRenderable;
import static it.tidalwave.bluebill.taxonomy.mobile.Taxon.Taxon;
import static it.tidalwave.semantic.document.Sendable.Sendable;

/***********************************************************************************************************************
 *
 * @stereotype Controller
 * 
 * @author  Fabrizio Giudici
 * @version $Id$
 *
 **********************************************************************************************************************/
public class DefaultObservationsViewController implements ObservationsViewController
  {
    private static final String CLASS = DefaultObservationsViewController.class.getName();
    private static final Logger logger = Logger.getLogger(CLASS);
    private static final Class<DefaultObservationsViewController> _ = DefaultObservationsViewController.class;

    @Nonnull
    private final ObservationClipboard observationClipboard = Locator.find(ObservationClipboard.class);

    @Nonnull
    private final LocationFinder locationFinder = Locator.find(LocationFinder.class);

    @Nonnull
    private final ObservationManager observationManager = Locator.find(ObservationManager.class);
    
    @Nonnull
    private final ObservationSet observationSet = observationManager.findOrCreateObservationSetById(null);
    
    @Nonnull
    private final Provider<SharingPreferences> sharingPreferences = Locator.createProviderFor(SharingPreferences.class);
    
    @Nonnull
    private final Provider<MasterFileSystem> masterFileSystem = Locator.createProviderFor(MasterFileSystem.class);

    @Nonnull
    private final ObservationsView view;

    /*******************************************************************************************************************
     *
     *
     *
     ******************************************************************************************************************/
    public DefaultObservationsViewController (final @Nonnull ObservationsView view)
      {
        this.view = view;
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    public void startNewObservationSequence()
      {
        ThreadAssertions.assertThread(ThreadType.UI);
        logger.info("startNewObservationSequence()");
        observationClipboard.clear();
        final Builder builder = observationClipboard.getBuilder();
        observationClipboard.setBuilder(builder.at(Calendar.getInstance().getTime()));
        locationFinder.start(); // search needs time, and we pre-start it
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    public void browseToFactSheet()
      {
        ThreadAssertions.assertThread(ThreadType.UI);
        logger.info("browseToFactSheet()");
        // nothing to do by default
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    public void commitNewObservation()
      {
        ThreadAssertions.assertThread(ThreadType.UI);
        logger.info("commitNewObservation()");
        
        final Observation.Builder builder = observationClipboard.getBuilder();

        if (builder.getItemCount() > 0)
          {
            builder.build();
            view.highlightLatestObservation();
            view.notifyNewObservationCommitted(getMessage(_, "addedNewObservation"));
          }

        observationClipboard.clear();
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    public void cancelNewObservation()
      {
        ThreadAssertions.assertThread(ThreadType.UI);
        logger.info("cancelNewObservation()");
        locationFinder.stop(); // we pre-started it
        view.notifyNewObservationCancelled(getMessage(_, "addedNewObservationCancelled"));
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    public void updateObservationDate (final @Nonnull Observation observation, final @Nonnull DateUpdater dateUpdater) 
      {
        ThreadAssertions.assertThread(ThreadType.UI);
        final Date date = dateUpdater.update(observation.getDate()); 
        observation.change().at(date).apply();
        view.notifyObservationDateTimeUpdated(getMessage(_, "observationDateTimeUpdated"));
      }
    
    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    public void exportObservations()
      {
        ThreadAssertions.assertThread(ThreadType.UI);
        logger.info("exportObservations()");
        view.askForExportOptions(new ReportQuestionWithFeedback(getMessage(_, "exportOptions"))                 
          {
            @Override
            protected void onConfirmInBackground()
              {
                logger.info("exportObservations(%s)", this);

                progressListener.start();
                final String reportBaseName = getReportBaseName();
                final Document report = getReportFactory().createReport(observationSet, reportBaseName, progressListener);
                progressListener.setProgressName(getMessage(_, "exportSaving"));
                
                final FileSystem fileSystem = masterFileSystem.get().getExternalFileSystem();
                final String reportExportFolderName = getMessage(_, "exportFolder");
                
                try
                  {
                    final File reportContainerFolder = fileSystem.getFile(reportExportFolderName);
                    final File reportFolder = new File(reportContainerFolder, reportBaseName);
                    logger.info(">>>> saving report to %s", reportFolder.getAbsolutePath());
                                                
                    if (!reportFolder.mkdirs() && !reportFolder.exists())
                      {
                        throw new IOException("Cannot create " + reportFolder.getAbsolutePath());   
                      }
                    
                    report.accept(new DocumentWriter(reportBaseName, reportFolder));
                    
                    // FIXME: this is brittle and based on a hidden behaviour of ExternalFileSystem.
                    final File storageRoot = reportContainerFolder.getParentFile().getParentFile();
                    final String path = reportFolder.getAbsolutePath().substring(storageRoot.getAbsolutePath().length());
                    final String title = getMessage(_, "export");
                    final String message = getMessage(_, "exportCompleted", path);
                    view.notifyExportCompleted(new QuestionWithFeedback(title, message));
                  }
                catch (Throwable t)
                  {
                    // FIXME: now that we have Exception reporting, why we just don't have this throwing exception?
                    logger.throwing(CLASS, "exportObservations()", t);
                    logger.severe("Unexpected exception while exporting: %s", t);
                    final String title = getMessage(_, "export");
                    final String message = getMessage(_, "exportFailed");
                    view.notifyExportFailed(new QuestionWithFeedback(title, message));
                    throw new RuntimeException(t);
                  }
                finally
                  {
                    progressListener.stop();
                  }            
              }
          });
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    public void shareObservations()
      {
        ThreadAssertions.assertThread(ThreadType.UI);
        logger.info("shareObservations()");
        view.askForShareOptions(new ReportQuestionWithFeedback(getMessage(_, "shareOptions"))
          {
            @Override
            protected void onConfirmInBackground()
              throws Exception
              {                
                try 
                  {
                    logger.info("shareObservations(%s)", this);
                    Document report = null;

                    try
                      {
                        progressListener.start();
                        final String reportBaseName = getReportBaseName();
                        report = getReportFactory().createReport(observationSet, reportBaseName, progressListener);
                      }
                    finally
                      {
                        progressListener.stop(); // before sending the report
                      }
                    
                    report.as(Sendable).sendTo(view, sharingPreferences.get().getDefaultRecipients());
//                    view.showLightNotification(NbBundle.getMessage(C, "shareCompleted"));
                    // TODO: a light notification that the operation completed, but we must get notified about the send activity.
                  }
                catch (Throwable t) 
                  {
                    // FIXME: now that we have Exception reporting, why we just don't have this throwing exception?
                    logger.throwing(CLASS, "shareObservations()", t);
                    logger.severe("Unexpected exception while sharing: %s", t);
                    final String title = getMessage(_, "share");
                    final String message = getMessage(_, "shareFailed");
                    view.notifyExportFailed(new QuestionWithFeedback(title, message));
                    throw new RuntimeException(t);
                  }
              }
          });
      }
    
    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    public void deleteAllObservations()
      {
        ThreadAssertions.assertThread(ThreadType.UI);
        logger.info("deleteAllObservations()");

        final String title = getMessage(_, "confirm");
        final String message = getMessage(_, "confirmEraseAllObservations");

        view.confirmToDeleteAllObservations(new QuestionWithFeedback(title, message)
          {
            @Override
            public void onConfirm() 
              {
                observationSet.clear();
                view.notifyAllObservationsDeleted(getMessage(_, "allObservationsErased"));
              }
          });
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    public void deleteObservationItem (final @Nonnull ObservationItem observationItem)
      {
        ThreadAssertions.assertThread(ThreadType.UI);
        logger.info("deleteObservationItem(%s)", observationItem);

        final Taxon taxon = observationItem.getObservable().as(Taxon);
        final String title = getMessage(_, "confirm");
        final String message = getMessage(_, "confirmDeleteObservationItem", taxon.as(StringRenderable).render());
        final String itemDeleted = getMessage(_, "itemDeleted");
        
        view.confirmToDeleteAnObservationItem(new QuestionWithFeedback(title, message)
          {
            @Override
            public void onConfirm() 
              {
                observationItem.getObservation().change().without(observationItem).apply();
                view.notifyObservationItemDeleted(itemDeleted);
              }
          });
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Nonnull
    private static String getReportBaseName()
      {
        final DateFormat df = new SimpleDateFormat("'blueBill-Observations-'yyyy-MM-dd_HHmm");
        return df.format(new Date());
      }
  }
