/***********************************************************************************************************************
 *
 * blueBill Mobile - Android - open source birding
 * Copyright (C) 2009-2011 by Tidalwave s.a.s. (http://www.tidalwave.it)
 *
 ***********************************************************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 *
 ***********************************************************************************************************************
 *
 * WWW: http://bluebill.tidalwave.it/mobile
 * SCM: https://kenai.com/hg/bluebill~android-src
 *
 **********************************************************************************************************************/
package it.tidalwave.bluebill.mobile.taxonomy.factsheet;

import javax.annotation.Nonnegative;
import javax.annotation.Nonnull;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.zip.ZipFile;
import java.util.zip.ZipEntry;
import java.io.File;
import java.io.InputStream;
import java.io.IOException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.openide.util.Lookup;
import org.openide.util.lookup.Lookups;
import org.openrdf.model.Graph;
import it.tidalwave.util.As;
import it.tidalwave.util.Id;
import it.tidalwave.util.thread.ThreadType;
import it.tidalwave.util.thread.annotation.ThreadConfined;
import it.tidalwave.role.ui.PresentationModel;
import it.tidalwave.netbeans.util.Locator;
import it.tidalwave.netbeans.capabilitiesprovider.CapabilitiesProvider;
import it.tidalwave.observation.ObservationSet;
import it.tidalwave.mobile.asset.SmartAssetManager;
import it.tidalwave.semantic.io.GraphDeserializer;
import it.tidalwave.semantic.io.GraphUnmarshaller;
import it.tidalwave.semantic.io.MimeTypes;
import it.tidalwave.semantic.io.impl.DefaultGraphDeserializer;
import it.tidalwave.semantic.io.impl.GraphUnmarshallerImpl;
import it.tidalwave.bluebill.taxonomy.mobile.Taxon;
import lombok.Cleanup;
import lombok.RequiredArgsConstructor;
import static it.tidalwave.role.Identifiable.Identifiable;

/***********************************************************************************************************************
 *
 * @author  Fabrizio Giudici
 * @version $Id$
 *
 **********************************************************************************************************************/
@RequiredArgsConstructor
public abstract class TaxonFactSheetViewControllerSupport<View extends TaxonFactSheetView> 
                                                           implements TaxonFactSheetViewController<View>, Lookup.Provider
  {
    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @RequiredArgsConstructor
    protected class RoleRegistrationBuilder
      {
        @Nonnull
        private final Object role;
        
        public void forClass (final @Nonnull Class<?> datumClass)
          {
            List<Object> roles = rolesMapByClass.get(datumClass);
            
            if (roles == null)
              {
                roles = new ArrayList<Object>();
                rolesMapByClass.put(datumClass, roles);
              }
            
            roles.add(role);
          }
      }

    private static final Logger log = LoggerFactory.getLogger(TaxonFactSheetViewControllerSupport.class);
    
    private static final Id TYPE_OBSERVATION_SET = new Id("http://www.tidalwave.it/rdf/observation/2010/07/01#ObservationSet"); // FIXME: put in vocabulary
    
    @Nonnull
    protected final View view;

    @Nonnull
    protected final Taxon taxon;

    private final List<PresentationModel> presentationModels = new ArrayList<PresentationModel>();

    private final Map<Class<?>, List<Object>> rolesMapByClass = new HashMap<Class<?>, List<Object>>();
    
    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Nonnull
    public PresentationModel getPresentationModel (final @Nonnegative int index)
      {
        return presentationModels.get(index);
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @ThreadConfined(type=ThreadType.ANY)
    public final void initialize()
      {
        log.info("initialize()");
        
        new Thread()
          {
            @Override
            public void run()
              {
                try 
                  {
                    clear();
                    createPresentationModels();
                  }
                catch (Throwable e)
                  {
                    log.error("While creating PresentationModels: {}", e.toString());
                    log.error("", e);
                    // FIXME: notify error on View, stop progress dialog
                  }
                    
                populate();
              }
          }.start();
      }    
    
    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    public void openWebPage (final @Nonnull As entity)
      {
        try
          {
            view.openWebPage(entity.as(Identifiable).getId().stringValue());
          }
        catch (Exception e)
          {
            log.warn("Can't open web page {}", e);
            log.warn("", e);
            // FIXME: non fatal notification to the view
          }
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    protected void clear()
      {
        presentationModels.clear();
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    protected void add (@Nonnull PresentationModel presentationModel)
      {
        presentationModels.add(presentationModel);
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @ThreadConfined(type=ThreadType.NOT_UI)
    protected abstract void createPresentationModels()
      throws Exception;
    
    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    protected void populate()
      {
        log.info("populate() - {} presentation models", presentationModels.size());
        view.populate(presentationModels);            
      }
    
    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Nonnull
    public ObservationSet loadObservationSet() // FIXME: should be protected, 
      throws IOException
      {
        // FIXME: now it loads xeno-canto, but should be a generic observation set with any kind of media
        log.info("loadObservationSet()");
        final String resourceName = "xenocanto-" + taxon.getId().stringValue().replaceAll(":", "_") + ".n3";
        log.info(">>>> resourceName: {}", resourceName);

        final File file = Locator.find(SmartAssetManager.class).findAsset("xeno-canto.zip");
        log.info(">>>> zip file: {}", file);
        final @Cleanup ZipFile zipFile = new ZipFile(file);
        final ZipEntry entry = zipFile.getEntry(resourceName);

        if (entry == null)
          {
            throw new IOException("No entry for " + resourceName);
          }

        final @Cleanup InputStream is = zipFile.getInputStream(entry);
        return loadObservationSet(is, MimeTypes.MIME_N3);
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Nonnull
    protected static ObservationSet loadObservationSet (final @Nonnull InputStream is, final @Nonnull String mimeType)
      throws IOException
      {
        try
          {
            final GraphDeserializer deserializer = new DefaultGraphDeserializer();
            final Graph graph = deserializer.read(is, mimeType);
            final GraphUnmarshaller unmarshaller = new GraphUnmarshallerImpl();
            return (ObservationSet)unmarshaller.unmarshal(graph, TYPE_OBSERVATION_SET);
          }
        catch (Exception e)
          {
            throw new RuntimeException(e);
//            throw new IOException(e.toString()); // Java 5 / Android compatibility
          }
      }    
    
    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Override @Nonnull
    public Lookup getLookup() 
      {
        final List<CapabilitiesProvider> capabilitiesProviders = new ArrayList<CapabilitiesProvider>();
        
        for (final Entry<Class<?>, List<Object>> entry : rolesMapByClass.entrySet())
          {
            final Class<?> datumClass = entry.getKey();
            final List<Object> roles = entry.getValue();
            
            capabilitiesProviders.add(new CapabilitiesProvider() 
              {
                @Nonnull
                public Class<?> getManagedClass() 
                  {
                    return datumClass;
                  } 

                @Nonnull
                public Collection<? extends Object> createCapabilities (final @Nonnull Object owner) 
                  {
                    return roles;
                  }

                @Nonnull
                public Collection<? extends Lookup> createLookups (final @Nonnull Object owner) 
                  {
                    return Collections.<Lookup>emptyList();
                  }                    
              });
          }
        
        return Lookups.fixed(capabilitiesProviders.toArray(new CapabilitiesProvider<?>[0]));
      }
    
    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Nonnull
    protected RoleRegistrationBuilder registerRole (final @Nonnull Object role)
      {
        return new RoleRegistrationBuilder(role);
      }
  }