/***********************************************************************************************************************
 *
 * blueBill Mobile - Android - open source birding
 * Copyright (C) 2009-2011 by Tidalwave s.a.s. (http://www.tidalwave.it)
 *
 ***********************************************************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 *
 ***********************************************************************************************************************
 *
 * WWW: http://bluebill.tidalwave.it/mobile
 * SCM: https://kenai.com/hg/bluebill~android-src
 *
 **********************************************************************************************************************/
package it.tidalwave.bluebill.mobile.taxonomy.factsheet.sound;

import javax.annotation.CheckForNull;
import javax.annotation.Nonnegative;
import javax.annotation.Nonnull;
import javax.inject.Provider;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.util.Arrays;
import java.util.Collection;
import java.util.List;
import java.io.IOException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import it.tidalwave.util.AsException;
import it.tidalwave.util.thread.ThreadType;
import it.tidalwave.util.thread.annotation.ThreadConfined;
import it.tidalwave.role.ui.ThreadBoundLookupContext;
import it.tidalwave.role.ui.tmpswing.Action;
import it.tidalwave.mobile.ui.ActionEvent;
import it.tidalwave.netbeans.util.Locator;
import it.tidalwave.observation.Observation;
import it.tidalwave.mobile.util.Downloadable;
import it.tidalwave.mobile.ui.QuestionWithFeedback;
import it.tidalwave.mobile.io.FileSystem;
import it.tidalwave.mobile.io.MasterFileSystem;
import it.tidalwave.mobile.media.MediaPlayer;
import it.tidalwave.mobile.media.MediaPlayer.Controller;
import it.tidalwave.mobile.media.Media;
import it.tidalwave.bluebill.taxonomy.mobile.Taxon;
import it.tidalwave.bluebill.mobile.taxonomy.factsheet.presentationmodel.MediaPresentationModel;
import it.tidalwave.bluebill.mobile.network.NetworkingPreferences;
import it.tidalwave.bluebill.mobile.taxonomy.factsheet.TaxonFactSheetViewControllerSupport;
import it.tidalwave.role.ui.spi.ActionProviderSupport;
import it.tidalwave.role.ui.tmpswing.AbstractAction;
import lombok.RequiredArgsConstructor;
import static org.openide.util.NbBundle.*;
import static it.tidalwave.role.Identifiable.Identifiable;
import static it.tidalwave.mobile.util.Downloadable.*;
import static it.tidalwave.mobile.util.Downloadable.Status.*;
import static it.tidalwave.mobile.media.Media.Media;
import static it.tidalwave.observation.Observation.Observation;

/***********************************************************************************************************************
 *
 * @stereotype Controller
 * 
 * @author  Fabrizio Giudici
 * @version $Id$
 *
 **********************************************************************************************************************/
public abstract class DefaultTaxonSoundFactSheetViewController extends TaxonFactSheetViewControllerSupport<TaxonSoundFactSheetView> 
                                                             implements TaxonSoundFactSheetViewController
  {
    private static final Logger log = LoggerFactory.getLogger(DefaultTaxonSoundFactSheetViewController.class);
    
    private static final Class<DefaultTaxonSoundFactSheetViewController> _ = DefaultTaxonSoundFactSheetViewController.class;

    private static final double MEGA = 1024.0 * 1024.0;
    /* package */ static final long MIN_FREE_SPACE = 20L * 1024 * 1024;
    private static final List<String> WATCHED_PROPERTIES = Arrays.asList(PROP_STATUS, PROP_DOWNLOAD_PROGRESS);

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    private enum Operation
      {
        DOWNLOAD
          {
            public void run (final @Nonnull Downloadable downloadable)
              {
                downloadable.download();
              }
          },
        REFRESH
          {
            public void run (final @Nonnull Downloadable downloadable)
              {
                downloadable.refresh();
              }
          };

        public abstract void run (@Nonnull Downloadable downloadable);
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @RequiredArgsConstructor
    public class DoNothingAction extends AbstractAction
      {
        public void actionPerformed (final @Nonnull ActionEvent event)
          {
          }      
      }
    
    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    public class DownloadAction extends AbstractAction
      {
        private final Media media;

        public DownloadAction (final @Nonnull Media media) 
          {
            super(getMessage(_, "downloadMedia"));
            this.media = media;
          }
        
        public void actionPerformed (final @Nonnull ActionEvent event)
          {
            downloadMedia(media);
          }

        @Override
        public boolean isEnabled() 
          {
            return isDownloadable(media);
          }
      }
    
    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    public class RefreshAction extends AbstractAction
      {
        private final Media media;
        
        public RefreshAction (final @Nonnull Media media) 
          {
            super(getMessage(_, "refreshMedia"));
            this.media = media;
          }
        
        public void actionPerformed (final @Nonnull ActionEvent event)
          {
            downloadMedia(media);
          }      

        @Override
        public boolean isEnabled() 
          {
            return isRefreshable(media);
          }
      }
    
    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    public class PlayAction extends AbstractAction
      {
        private final Media media;
        
        public PlayAction (final @Nonnull Media media) 
          {
            super(getMessage(_, "playMedia"));
            this.media = media;
          }
        
        public void actionPerformed (final @Nonnull ActionEvent event)
          {
            playMedia(media);
          }      
        
        @Override
        public boolean isEnabled() 
          {
            return isPlayable(media);
          }
      }
    
    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    public class StopAction extends AbstractAction
      {
        private final Media media;
        
        public StopAction (final @Nonnull Media media) 
          {
            super("stop");
            this.media = media;
          }
        
        public void actionPerformed (final @Nonnull ActionEvent event)
          {
            stopPlayer();
          }      
        
        @Override
        public boolean isEnabled() 
          {
            return isPlaying(media);
          }
      }
    
    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    public class DeleteAction extends AbstractAction
      {
        private final Media media;
        
        public DeleteAction (final @Nonnull Media media) 
          {
            super(getMessage(_, "deleteMedia"));
            this.media = media;
          }
        
        public void actionPerformed (final @Nonnull ActionEvent event)
          {
            deleteMedia(media);
          }      
        
        @Override
        public boolean isEnabled() 
          {
            return isDeletable(media);
          }
      }
    
    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    public class OpenWebPageAction extends AbstractAction
      {
        private final Observation observation;
        
        public OpenWebPageAction (final @Nonnull Observation observation) 
          {
            super(getMessage(_, "openMediaWebPage"));
            this.observation = observation;
          }
        
        public void actionPerformed (final @Nonnull ActionEvent event)
          {
            openWebPage(observation);
          }      
        
        @Override
        public boolean isEnabled() 
          {
            return hasWebPage(observation);
          }
      }
    
    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    class SoundActionProvider extends ActionProviderSupport
      {
        @Nonnull
        private final Media media;
        
        private final Action doNothingAction = new DoNothingAction();
        private final Action downloadAction;
        private final Action refreshAction;
        private final Action playAction;
        private final Action stopAction;
        private final Action deleteAction;
        private final Action openWebPageAction;

        public SoundActionProvider (final @Nonnull Media media, final @Nonnull Observation observation)
          {
            this.media = media;
            downloadAction = new DownloadAction(media);
            refreshAction = new RefreshAction(media);
            playAction = new PlayAction(media);
            stopAction = new StopAction(media);
            deleteAction = new DeleteAction(media);
            openWebPageAction = new OpenWebPageAction(observation);
          }

        @Override @Nonnull
        public Collection<? extends Action> getActions()
          {
            return Arrays.asList(playAction, downloadAction, refreshAction, deleteAction, openWebPageAction);
          }
        
        @Override @Nonnull
        public Action getDefaultAction()
          {
            if (isPlaying(media))
              {
                return stopAction;
              }
            else if (isPlayable(media))
              {
                return playAction;
              }
            else if (isDownloadable(media))
              {
                return downloadAction;
              }
            else
              {
                return doNothingAction;
              }
          }
      }
    
    @CheckForNull
    /* package */ Controller mediaPlayerController;

    @Nonnull
    private final Provider<MediaPlayer> mediaPlayer = Locator.createProviderFor(MediaPlayer.class);

    @Nonnull
    private final Provider<MasterFileSystem> masterFileSystem = Locator.createProviderFor(MasterFileSystem.class);

    @Nonnull
    private final Provider<NetworkingPreferences> networkingPreferences = Locator.createProviderFor(NetworkingPreferences.class);

    @CheckForNull
    private Media playingMedia;

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    /* package */ final PropertyChangeListener mediaChangeListener = new PropertyChangeListener()
      {
        public void propertyChange (final @Nonnull PropertyChangeEvent event)
          {
            notifyStatusChanged();
          }
      };

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    /* package */ final PropertyChangeListener downloadablePropertyChangeListener = new PropertyChangeListener()
      {
        public void propertyChange (final @Nonnull PropertyChangeEvent event)
          {
//            logger.info("propertyChange(%s)", event);

            if (WATCHED_PROPERTIES.contains(event.getPropertyName()) && (event.getOldValue() != event.getNewValue()))
              {
                notifyStatusChanged();

                if (event.getPropertyName().equals(PROP_STATUS))
                  {
                    final Downloadable downloadable = (Downloadable)event.getSource();
                    
                    switch ((Status)event.getNewValue())
                      {
                        case DOWNLOADED:
                          final double freeSpace = masterFileSystem.get().getExternalFileSystem().getFreeSpace() / MEGA;
                          view.notifyCompletion(getMessage(_, "downloadComplete", freeSpace));
                          downloadable.removePropertyChangeListener(downloadablePropertyChangeListener);
                          break;

                        case BROKEN:
                          view.notifyError(getMessage(_, "downloadFailed"));
                          downloadable.removePropertyChangeListener(downloadablePropertyChangeListener);
                          break;

                        case DOWNLOADING:
                        case NOT_DOWNLOADED:
                        case OBSOLETE:
                        case QUEUED:
                      }
                  }
              }
          }
      };

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    public DefaultTaxonSoundFactSheetViewController (final @Nonnull TaxonSoundFactSheetView view, 
                                                     final @Nonnull Taxon taxon)
      {
        super(view, taxon);
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Override
    public void disposePlayer()
      {
        log.info("disposePlayer()");
        
        if (mediaPlayerController != null)
          {
            mediaPlayerController.removePropertyChangeListener(mediaChangeListener);
            mediaPlayerController.dispose();
            mediaPlayerController = null;
          }
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    boolean isDownloadable (final @Nonnull Media media)
      {
        return isStatus(media, NOT_DOWNLOADED, OBSOLETE, BROKEN);
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    boolean isRefreshable (final @Nonnull Media media)
      {
        return isStatus(media, DOWNLOADED, OBSOLETE, BROKEN);
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    boolean isDeletable (final @Nonnull Media media)
      {
        return isStatus(media, DOWNLOADED, OBSOLETE);
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    boolean isPlayable (final @Nonnull Media media)
      {
        return isStatus(media, DOWNLOADED, OBSOLETE);
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    boolean isPlaying (final @Nonnull Media media)
      {
        return isPlayable(media) && (mediaPlayerController != null) && mediaPlayerController.isPlaying() && (media == playingMedia);
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    boolean hasWebPage (final @Nonnull Observation observation)
      {
        try
          {
            observation.as(Identifiable).getId(); // to probe existence of id
            return true;
          }
        catch (AsException e)
          {
            return false;
          }
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    void downloadMedia (final @Nonnull Media media)
      {
        log.info("downloadMedia({})", media);
        
        if (isDownloadable(media))
          {
            downloadMediaWithConfirmation(media, Operation.DOWNLOAD);
          }
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    void refreshMedia (final @Nonnull Media media)
      {
        log.info("refreshMedia({})", media);
        
        if (isRefreshable(media))
          {
            downloadMediaWithConfirmation(media, Operation.REFRESH);
          }
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    void deleteMedia (final @Nonnull Media media)
      {
        log.info("deleteMedia({})", media);
        
        if (isDeletable(media))
          {
            final String title   = getMessage(_, "confirmDeletionTitle");
            final String message = getMessage(_, "confirmDeletionMessage");

            view.confirmToDeleteMedia(new QuestionWithFeedback(title, message)
              {
                @Override
                public void onConfirm()
                  {
                    try
                      {
                        media.as(Removable).remove();
                        final double freeSpace = masterFileSystem.get().getExternalFileSystem().getFreeSpace() / MEGA;
                        view.notifyCompletion(getMessage(_, "deletionCompleted", freeSpace));
                      }
                    catch (Exception e)
                      {
                        log.warn("Can't delete sound {}", e);
                        log.warn("", e);
                        view.notifyError(getMessage(_, "deletionFailed"));
                      }

                    // Needed because at this point the PropertyChangeListener has been detached.
                    // Maybe it would be better to keep the listener attached, implement a dispose() method that
                    // detaches it at the end.
                    notifyStatusChanged();
                  }
              });
          }
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    void playMedia (final @Nonnull Media media)
      {
        log.info("playMedia({})", media);
        
        if (isPlayable(media))
          {
            try
              {
                disposePlayer();
                playingMedia = media;
                mediaPlayerController = mediaPlayer.get().play(media);
                mediaPlayerController.addPropertyChangeListener(mediaChangeListener);
                view.notifyMediaPlaying(mediaPlayerController);
                notifyStatusChanged();
              }
            catch (Exception e)
              {
                log.warn("Can't play sound {}", e);
                log.warn("", e);
                view.notifyError(getMessage(_, "playFailed"));
              }
          }
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    void stopPlayer()
      {
        log.info("stopPlayer()");
        
        disposePlayer();
        notifyStatusChanged();
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    Media getMedia (final @Nonnegative int index)
      {
        return getObservation(index).as(Media);
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    Observation getObservation (final @Nonnegative int index)
      {
        return getPresentationModel(index).as(Observation);
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @ThreadConfined(type=ThreadType.NOT_UI)
    protected void createPresentationModels()
      throws IOException 
      {
        ThreadBoundLookupContext.with(this)
                      .run(new ThreadBoundLookupContext.Task<IOException>() 
          {
            public void run() 
              throws IOException
              {
                for (final Observation observation : loadObservationSet().find(Observation).results())
                  {
                    final Media media = observation.as(Media);
                    add(new MediaPresentationModel(media, observation, new SoundActionProvider(media, observation)));
                  }
              }
          });          
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    protected void notifyStatusChanged()
      {
        view.notifyStatusChanged();  
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    private void downloadMediaWithConfirmation (final @Nonnull Media media, final @Nonnull Operation operation)
      {
        final String title   = getMessage(_, "confirmDownloadTitle");
        final String message = getMessage(_, "confirmDownloadMessage");

        if (networkingPreferences.get().isNetworkConnectionAllowed())
          {
            doDownloadMedia(media, operation);
          }
        else
          {
            view.confirmToDownloadMedia(new QuestionWithFeedback(title, message)
              {
                @Override
                public void onConfirm()
                  {
                    doDownloadMedia(media, operation);
                  }
              });
          }
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    /* package */ void doDownloadMedia (final @Nonnull Media media, final @Nonnull Operation operation)
      {          
        final FileSystem externalFileSystem = masterFileSystem.get().getExternalFileSystem();
        
        if (!externalFileSystem.isWritable())
          {
            final String title   = getMessage(_, "downloadFailed");
            final String message = getMessage(_, "diskNotMountedMessage");
            view.notifySdCardNotReady(new QuestionWithFeedback(title, message));
          }
        else if (externalFileSystem.getFreeSpace() <= MIN_FREE_SPACE)
          {
            final String title   = getMessage(_, "downloadFailed");
            final String message = getMessage(_, "diskFullMessage", MIN_FREE_SPACE / MEGA);
            view.notifySdCardNotReady(new QuestionWithFeedback(title, message));
          }
        else
          {
            final Downloadable downloadable = media.as(Downloadable);
            downloadable.addPropertyChangeListener(downloadablePropertyChangeListener);
            operation.run(downloadable);
            ping();
          }
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    private boolean isStatus (final @Nonnull Media media, final @Nonnull Status ... status)
      {
        final Downloadable downloadable = media.as(Downloadable);
        return Arrays.asList(status).contains(downloadable.getStatus());
      }

    protected void ping()
      {
      }
  }
